package cz.uhk.fim.smarttodo.domain;

import cz.uhk.fim.smarttodo.dto.TodoItemResponseDto;
import cz.uhk.fim.smarttodo.dto.todo.TodoItemDto;
import org.hibernate.annotations.CreationTimestamp;
import org.hibernate.annotations.OnDelete;
import org.hibernate.annotations.OnDeleteAction;

import javax.persistence.*;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.List;
import java.util.stream.Collectors;

@Entity
@Table(name = "\"TODO_ITEM\"")
public class TodoItem {

    public enum TodoType {
        TODO,
        NESTED_TODO;

        TodoType() {
        }

    }

    @Id
    @GeneratedValue(strategy = GenerationType.IDENTITY)
    private Integer todoId;

    @Column
    private String name;

    @Column(nullable = false)
    private boolean done = false;

    @Column(nullable = false)
    private boolean flagged = false;

    @Column
    private Date onDate;

    @OneToMany(cascade = {
            CascadeType.ALL
    }, orphanRemoval = true)
    @JoinTable(name = "NESTED_TODO_ITEM")
    @OrderBy("todoId ASC")
    private List<TodoItem> nestedTodoItems = new ArrayList<>();

    @OneToOne(mappedBy = "todoItem", cascade = CascadeType.ALL, orphanRemoval = true)
    private Place place;

    @Enumerated(EnumType.STRING)
    private TodoType todoType;

    @CreationTimestamp
    private Date createdOn;

    @ManyToOne
    @JoinColumn(name = "list_Id", nullable = false)
    private TodoList todoList;

    public static TodoItem createTodoItem(TodoItemDto todoItemDto, TodoList todoList) {
        TodoItem todoItem = mapTodoItem(todoItemDto.getName(), todoList, false);

        if (todoItemDto.getNestedTodoItems() != null) {
            List<TodoItem> nestedTodos = Arrays
                    .stream(todoItemDto.getNestedTodoItems())
                    .map(todo -> mapTodoItem(todo.getName(), todoList, true))
                    .collect(Collectors.toList());

            todoItem.setNestedTodoItems(nestedTodos);
        }

        return todoItem;
    }

    private static TodoItem mapTodoItem(String name, TodoList todoList, boolean nested) {
        TodoItem todoItem = new TodoItem();
        todoItem.setDone(false);
        todoItem.setFlagged(false);
        todoItem.setName(name);
        todoItem.setTodoType(nested ? TodoType.NESTED_TODO : TodoType.TODO);
        todoItem.setTodoList(todoList);

        return todoItem;
    }

    public TodoItemResponseDto mapTodoItemToResponse() {
        TodoItemResponseDto todoDto = new TodoItemResponseDto();
        todoDto.setTodoId(this.getTodoId());
        todoDto.setName(this.getName());
        todoDto.setCreatedOn(this.getCreatedOn());
        todoDto.setTodoType(this.getTodoType());
        todoDto.setDone(this.isDone());
        todoDto.setFlagged(this.isFlagged());
        todoDto.setOnDate(this.getOnDate());
        todoDto.setListId(this.getTodoListId());
        todoDto.setPlace(this.getPlace() != null ? this.getPlace().mapPlaceToResponse() : null);

        if (!this.isNested()) {
            List<TodoItemResponseDto> nestedTodoItems = new ArrayList<>();
            for (TodoItem nestedTodoItem : this.getNestedTodoItems()) {
                TodoItemResponseDto nestedTodoDto = new TodoItemResponseDto();
                nestedTodoDto.setTodoId(nestedTodoItem.getTodoId());
                nestedTodoDto.setName(nestedTodoItem.getName());
                nestedTodoDto.setCreatedOn(nestedTodoItem.getCreatedOn());
                nestedTodoDto.setTodoType(nestedTodoItem.getTodoType());
                nestedTodoDto.setNestedTodoItems(null);
                nestedTodoDto.setDone(nestedTodoItem.isDone());
                nestedTodoDto.setFlagged(nestedTodoItem.isFlagged());
                nestedTodoDto.setOnDate(nestedTodoItem.getOnDate());
                nestedTodoDto.setListId(this.getTodoListId());
                nestedTodoDto.setPlace(nestedTodoItem.getPlace() != null ? nestedTodoItem.getPlace().mapPlaceToResponse() : null);

                nestedTodoItems.add(nestedTodoDto);
            }

            todoDto.setNestedTodoItems(nestedTodoItems);
        }

        return todoDto;
    }

    public void changeTodoList(TodoList todoList) {
        this.setTodoList(todoList);

        if (this.getTodoType().equals(TodoType.TODO))
            this.nestedTodoItems
                    .forEach(todo -> todo.setTodoList(todoList));
        else
            this.setTodoType(TodoType.TODO);
    }

    public void editNestecTodosByRequest(TodoItemDto.NestedTodoDto[] nestedTodoItems) {
        if (nestedTodoItems.length == 0) {
            this.getNestedTodoItems().clear();
            return;
        }

        if (this.getNestedTodoItems().isEmpty()) {
            List<TodoItem> nestedTodos = new ArrayList<>();

            for (TodoItemDto.NestedTodoDto nestedTodoItem : nestedTodoItems)
                nestedTodos.add(mapTodoItem(nestedTodoItem.getName(), this.todoList, true));

            this.getNestedTodoItems().addAll(nestedTodos);
            return;
        }


        List<TodoItem> editedNestedTodos = new ArrayList<>();

        for (TodoItemDto.NestedTodoDto nestedTodoItem : nestedTodoItems) {

            if (nestedTodoItem.getTodoId() == null)
                editedNestedTodos.add(mapTodoItem(nestedTodoItem.getName(), this.todoList, true));
            else {
                this.getNestedTodoItems()
                        .stream()
                        .filter(todo -> nestedTodoItem.getTodoId().equals(todo.getTodoId()))
                        .findFirst()
                        .ifPresent(editedNestedTodos::add);
            }
        }

        this.getNestedTodoItems().clear();
        this.getNestedTodoItems().addAll(editedNestedTodos);

    }


    public Integer getTodoId() {
        return todoId;
    }

    public void setTodoId(Integer todoId) {
        this.todoId = todoId;
    }

    public String getName() {
        return name;
    }

    public void setName(String name) {
        this.name = name;
    }

    public List<TodoItem> getNestedTodoItems() {
        return nestedTodoItems;
    }

    public void setNestedTodoItems(List<TodoItem> nestedTodoItems) {
        this.nestedTodoItems = nestedTodoItems;
    }

    public TodoType getTodoType() {
        return todoType;
    }

    public void setTodoType(TodoType todoType) {
        this.todoType = todoType;
    }

    public Date getCreatedOn() {
        return createdOn;
    }

    public void setCreatedOn(Date createdOn) {
        this.createdOn = createdOn;
    }

    public TodoList getTodoList() {
        return todoList;
    }

    public void setTodoList(TodoList todoList) {
        this.todoList = todoList;
    }

    public boolean isDone() {
        return done;
    }

    public void setDone(boolean done) {
        this.done = done;
    }

    public boolean isFlagged() {
        return flagged;
    }

    public void setFlagged(boolean flagged) {
        this.flagged = flagged;
    }

    public Date getOnDate() {
        return onDate;
    }

    public void setOnDate(Date onDate) {
        this.onDate = onDate;
    }

    public Place getPlace() {
        return place;
    }

    public void setPlace(Place place) {
        this.place = place;
    }

    public Integer getTodoListId() {
        return this.todoList.getListId();
    }

    public boolean isNested() {
        return this.todoType.equals(TodoType.NESTED_TODO);
    }
}
