package cz.uhk.fim.smarttodo.service;

import cz.uhk.fim.smarttodo.common.Constant;
import cz.uhk.fim.smarttodo.domain.Place;
import cz.uhk.fim.smarttodo.domain.TodoItem;
import cz.uhk.fim.smarttodo.domain.TodoList;
import cz.uhk.fim.smarttodo.domain.User;
import cz.uhk.fim.smarttodo.dto.NewTodoListRequestDto;
import cz.uhk.fim.smarttodo.dto.TodoListResponseDto;
import cz.uhk.fim.smarttodo.dto.todo.TodoItemDto;
import cz.uhk.fim.smarttodo.dto.user.UserForListEditDto;
import cz.uhk.fim.smarttodo.dto.witAi.ReceivedMessageDto;
import cz.uhk.fim.smarttodo.repository.PlaceRepository;
import cz.uhk.fim.smarttodo.repository.TodoItemRepository;
import cz.uhk.fim.smarttodo.repository.TodoListRepository;
import cz.uhk.fim.smarttodo.repository.UserRepository;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.*;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.web.client.RestClientException;
import org.springframework.web.client.RestTemplate;

import javax.persistence.EntityManager;
import javax.persistence.PersistenceContext;
import java.util.*;
import java.util.stream.Collectors;

@Service
public class TodoService {

    @PersistenceContext
    private EntityManager em;

    private final TodoItemRepository todoItemRepository;
    private final TodoListRepository todoListRepository;
    private final PlaceRepository placeRepository;
    private final UserRepository userRepository;

    @Autowired
    public TodoService(TodoItemRepository todoItemRepository, TodoListRepository todoListRepository, UserRepository userRepository, PlaceRepository placeRepository) {
        this.todoItemRepository = todoItemRepository;
        this.todoListRepository = todoListRepository;
        this.userRepository = userRepository;
        this.placeRepository = placeRepository;
    }

    private static final String WIT_AI_EN_KEY = "L5C7NUMGNUB2PUMMT6VQLTJCYNEOLFSZ";
    private static final String WIT_AI_CS_KEY = "QVEXXHHQM6IURIY37BYPMFXSPW2EI3GN";
    private static final String WIT_AI_MESSAGE_PATH = "https://api.wit.ai/message";
    private static final String WIT_AI_VERSION = "20200423";

    public List<TodoListResponseDto> getAllUsersTodoLists(User user) {

        return user.getAllUsersTodoLists();
    }

    private boolean isUsersTodoList(User user, TodoList todoList) {
        return isUsersTodoList(user.getId(), todoList);
    }

    private boolean isUsersTodoList(Integer userId, TodoList todoList) {
        return todoList.getUsers().stream().anyMatch(todoUser -> todoUser.getId().equals(userId));
    }

    @Transactional
    public void createOrUpdateTodoList(NewTodoListRequestDto dto, User user) {
        if (dto.getListId() != null) {
            Optional<TodoList> list = todoListRepository.findById(dto.getListId());

            if (list.isPresent()) {
                TodoList todoList = list.get();

                if (!isUsersTodoList(user, todoList))
                    return;

                todoList.setName(dto.getName());
                todoList.setColor(dto.getColor());
                todoList.setIcon(dto.getIcon());
            }
        } else {
            TodoList todoList = new TodoList();
            todoList.setName(dto.getName());
            todoList.setColor(dto.getColor());
            todoList.setIcon(dto.getIcon());
            todoList.setReferralCode(UUID.randomUUID().toString());
            todoList.setListOwner(user);

            user.getTodoLists().add(todoList);

            userRepository.save(user);
        }

    }

    public void deleteTodoList(User user, Integer todoListId) {
        deleteTodoListFromUser(user, todoListId);
        deleteTodoListIfHasNoUsers(todoListId);
    }

    public void deleteTodoListFromUser(User user, Integer todoListId) {
        Optional<TodoList> list = todoListRepository.findById(todoListId);

        if (list.isPresent()) {
            TodoList todoList = list.get();

            if (isUsersTodoList(user, todoList)) {

                if (todoList.getUsers().size() > 1 && todoList.getListOwner().getId().equals(user.getId())) {
                    Optional<User> potentialOwner = todoList.getUsers()
                            .stream()
                            .filter(u -> !u.getId().equals(user.getId()))
                            .findFirst();

                    potentialOwner.ifPresent(user1 -> {
                        todoList.setListOwner(user1);
                        todoListRepository.save(todoList);
                    });
                }

                user.removeTodoList(todoList);
                userRepository.save(user);
            }

        }

    }

    private void deleteTodoListIfHasNoUsers(Integer todoListId) {
        Optional<TodoList> list = todoListRepository.findById(todoListId);

        if (list.isPresent()) {
            TodoList todoList = list.get();

            if (todoList.getUsers().isEmpty())
                todoListRepository.delete(todoList);

        }
    }

    public TodoItem saveTodo(User user, TodoItemDto todoItemDto) {
        Optional<TodoList> list = todoListRepository.findById(todoItemDto.getListId());

        if (list.isPresent()) {
            TodoList todoList = list.get();

            TodoItem todoItem = TodoItem.createTodoItem(todoItemDto, todoList);

            todoItemRepository.save(todoItem);

            return todoItem;
        }
        return null;
    }

    public void deleteTodo(User user, Integer todoId) {
        Optional<TodoItem> todo = todoItemRepository.findById(todoId);

        if (todo.isPresent()) {
            TodoItem todoItem = todo.get();

            if (user.hasAccessToTodoList(todoItem.getTodoList().getListId())) {
                if (todoItem.getTodoType().equals(TodoItem.TodoType.NESTED_TODO)) {
                    deleteNestedTodo(todoItem);

                } else
                    todoItemRepository.delete(todoItem);
            }
        }
    }

    public TodoItem editTodo(User user, TodoItemDto todoItemDto) {
        Optional<TodoItem> todo = todoItemRepository.findById(todoItemDto.getTodoId());

        if (todo.isPresent()) {

            TodoItem todoItem = todo.get();

            if (!(user.hasAccessToTodoList(todoItemDto.getListId()) && user.hasAccessToTodoList(todoItem.getTodoListId())))
                return null;

            Optional<TodoList> todoList = todoListRepository.findById(todoItemDto.getListId());

            todoItem.setName(todoItemDto.getName());
            todoItem.setFlagged(todoItemDto.isFlagged());
            todoItem.setDone(todoItemDto.isDone());
            todoItem.setOnDate(todoItemDto.getOnDate());
            Place place = todoItem.getPlace();
            if (todoItemDto.getPlace() != null) {

                if (place != null) {
                    place.setLat(todoItemDto.getPlace().getLat());
                    place.setLng(todoItemDto.getPlace().getLng());
                } else {
                    place = new Place();
                    place.setLat(todoItemDto.getPlace().getLat());
                    place.setLng(todoItemDto.getPlace().getLng());
                    place.setTodoItem(todoItem);
                }

                todoItem.setPlace(place);
            } else {

                todoItem.setPlace(null);
            }

            if (todoItem.getTodoType().equals(TodoItem.TodoType.TODO))
                todoItem.editNestecTodosByRequest(todoItemDto.getNestedTodoItems());

            if (todoList.isPresent())
                if (!todoItem.getTodoList().getListId().equals(todoList.get().getListId())) {
                    if (todoItem.getTodoType().equals(TodoItem.TodoType.NESTED_TODO)) {
                        deleteNestedTodo(todoItem);
                    }
                    todoItem.changeTodoList(todoList.get());

                }

            return todoItemRepository.save(todoItem);
        }
        return null;
    }

    private void deleteNestedTodo(TodoItem todoItem) {
        todoItem.getTodoList().getItems()
                .stream()
                .filter(parentTodo -> {
                    for (TodoItem nestedTodoItem : parentTodo.getNestedTodoItems()) {
                        if (nestedTodoItem.getTodoId().equals(todoItem.getTodoId()))
                            return true;
                    }
                    return false;
                })
                .findFirst()
                .ifPresent(parentTodoItem -> {
                    parentTodoItem.getNestedTodoItems().remove(todoItem);
                    todoItemRepository.save(parentTodoItem);
                });
    }

    public void toggleTodo(User user, Integer todoId) {
        Optional<TodoItem> todo = todoItemRepository.findById(todoId);

        if (todo.isPresent()) {
            TodoItem todoItem = todo.get();

            if (user.hasAccessToTodoList(todoItem.getTodoList().getListId())) {
                todoItem.setDone(!todoItem.isDone());

                if (!todoItem.isNested() && todoItem.isDone())
                    todoItem.getNestedTodoItems()
                            .forEach(nestedTodo -> nestedTodo.setDone(true));

                todoItemRepository.save(todoItem);
            }
        }
    }

    public void addNewUserToTodoList(String referralCode, User user) {
        TodoList todoList = todoListRepository.findByReferralCode(referralCode);

        if (todoList != null) {
            boolean userHasList = user.getTodoLists()
                    .stream()
                    .anyMatch(list -> list.getReferralCode().equals(todoList.getReferralCode()));

            if (!userHasList) {
                user.getTodoLists().add(todoList);
                userRepository.save(user);
            }
        }

    }

    public List<TodoItem> findTodosByText(User user, String text, Locale locale) {
        ReceivedMessageDto processedText = processUsersText(text, locale);
        if (processedText == null)
            return new ArrayList<>();

        return filterRequestedTodos(user.getTodoLists(), processedText);
    }

    private List<TodoItem> filterRequestedTodos(List<TodoList> todoLists, ReceivedMessageDto processedText) {
        List<TodoItem> filteredTodos = new ArrayList<>();

        ReceivedMessageDto.Entity entities = processedText.getEntities();

        if (entities == null)
            return filteredTodos;

        // Intent entity process
        ReceivedMessageDto.Entity.Intent[] intent = entities.getIntent();
        if (intent != null) {
            for (ReceivedMessageDto.Entity.Intent inten : intent) {
                if (inten.getValue().equals(Constant.ALL)) {
                    for (TodoList todoList : todoLists) {
                        filteredTodos.addAll(todoList.getItems());
                    }

                    return filteredTodos;
                }
            }
        }

        // datetime entity process
        ReceivedMessageDto.Entity.Datetime[] dateFilters = entities.getDatetime();
        if (dateFilters != null)
            for (ReceivedMessageDto.Entity.Datetime dateFilter : dateFilters) {
                if (dateFilter.getValue() != null) {
                    for (TodoList todoList : todoLists)
                        filteredTodos.addAll(todoList.getTodosForDate(dateFilter.getValue()));

                } else if (dateFilter.getFrom() != null && dateFilter.getTo() == null) {
                    for (TodoList todoList : todoLists)
                        filteredTodos.addAll(todoList.getTodosFromDate(dateFilter.getFrom().getValue()));

                } else if (dateFilter.getFrom() == null && dateFilter.getTo() != null) {
                    for (TodoList todoList : todoLists)
                        filteredTodos.addAll(todoList.getTodosToDate(dateFilter.getTo().getValue()));

                } else if (dateFilter.getFrom() != null && dateFilter.getTo() != null) {
                    for (TodoList todoList : todoLists)
                        filteredTodos.addAll(todoList.getTodosBetweenDate(dateFilter.getFrom().getValue(), dateFilter.getTo().getValue()));

                }
            }

        return filteredTodos;
    }

    private ReceivedMessageDto processUsersText(String text, Locale locale) {
        ReceivedMessageDto dto = null;
        switch (locale.getLanguage()) {
            case "cs":
                dto = processCsText(text);
                break;
            default:
            case "en":
                dto = processEnText(text);
                break;
        }

        return dto;
    }

    private ReceivedMessageDto processCsText(String text) {
        ReceivedMessageDto dto = processTextOnWitAiSide(text, WIT_AI_CS_KEY);
        if (dto == null) return null;
        final String possibleEnSentence = dto.getPossibleEnSentence(text);
        System.out.println(possibleEnSentence);

        return processEnText(possibleEnSentence);
    }

    private ReceivedMessageDto processEnText(String text) {
        return processTextOnWitAiSide(text, WIT_AI_EN_KEY);
    }

    private ReceivedMessageDto processTextOnWitAiSide(String text, String witAiKey) {
        HttpHeaders headers = getWitAiAuthorizedHeaders(witAiKey);
        HttpEntity<String> httpEntity = new HttpEntity<>(null, headers);

        RestTemplate restTemplate = new RestTemplate();
        ResponseEntity<ReceivedMessageDto> exchange = null;
        try {
            exchange = restTemplate.exchange(String.format("%s?v=%s&q=%s", WIT_AI_MESSAGE_PATH, WIT_AI_VERSION, text), HttpMethod.GET, httpEntity, ReceivedMessageDto.class);
        } catch (RestClientException e) {
            return null;
        }
        if (exchange.getStatusCode().equals(HttpStatus.OK))
            return exchange.getBody();
        else
            return null;
    }

    private HttpHeaders getWitAiAuthorizedHeaders(String key) {
        HttpHeaders headers = new HttpHeaders();
        headers.add(Constant.AUTHORIZATION_HEADER, String.format("%s%s", Constant.TOKEN_PREFIX, key));
        headers.setContentType(MediaType.APPLICATION_JSON);

        return headers;
    }


    public void deleteWholeList(User user, Integer todoListId) {
        Optional<TodoList> list = todoListRepository.findById(todoListId);

        if (list.isPresent()) {
            TodoList todoList = list.get();

            if (todoList.getListOwner().getId().equals(user.getId())) {

                todoList.clearList();
                todoListRepository.delete(todoList);
            }

        }
    }

    public void removeUserFromList(User owner, Integer userId, Integer listId) {
        Optional<TodoList> list = todoListRepository.findById(listId);

        if (list.isPresent()) {
            TodoList todoList = list.get();

            if (todoList.getListOwner().getId().equals(owner.getId())) {
                todoList.removeUserFromList(userId);

                todoListRepository.save(todoList);
            }
        }
    }

    public void changeListOwner(User owner, Integer userId, Integer listId) {
        Optional<TodoList> list = todoListRepository.findById(listId);

        if (list.isPresent()) {
            TodoList todoList = list.get();

            if (todoList.getListOwner().getId().equals(owner.getId())) {
                Optional<User> newOwner = userRepository.findById(userId);

                if (newOwner.isPresent()) {
                    todoList.setListOwner(newOwner.get());

                    todoListRepository.save(todoList);
                }
            }

        }
    }

    public List<UserForListEditDto> getAllListUsers(User user, Integer listId) {
        Optional<TodoList> list = todoListRepository.findById(listId);

        if (list.isPresent()) {
            TodoList todoList = list.get();

            if (todoList.getListOwner().getId().equals(user.getId())) {

                return todoList.getUsers()
                        .stream()
                        .filter(u -> !u.getId().equals(user.getId()))
                        .map(u -> {
                            UserForListEditDto dto = new UserForListEditDto();
                            dto.setUserId(u.getId());
                            dto.setUsername(u.getUsername());
                            dto.setInitials(u.getFirstName() != null && u.getLastName() != null ? String.format("%s%s", u.getFirstName().charAt(0), u.getLastName().charAt(0)) : null);

                            return dto;
                        })
                        .collect(Collectors.toList());
            }
        }
        return null;
    }
}
