import AxiosCommon from "../../axios/axios-common";
import moment from "moment";

const staticEventListTypes = Object.freeze({
  today: "TODAY",
  tomorrow: "TOMORROW",
  others: "OTHERS"
});

const state = {
  eventList: [],
  eventsRefreshed: false
};

const getters = {
  getTodayEvents: state => {
    const today = Date.now();
    return state.eventList.filter(e => moment(today).isSame(e.onDate, "day"));
  },
  getTomorrowEvents: state => {
    const tomorrow = new Date();
    tomorrow.setDate(tomorrow.getDate() + 1);

    return state.eventList.filter(e =>
      moment(tomorrow).isSame(e.onDate, "day")
    );
  },
  getOtherEvents: state => {
    const today = Date.now();
    const tomorrow = new Date();
    tomorrow.setDate(tomorrow.getDate() + 1);

    return state.eventList
      .filter(
        e =>
          !moment(tomorrow).isSame(e.onDate, "day") &&
          moment(today).isBefore(e.onDate, "day")
      )
      .sort((a, b) => new Date(a.onDate) - new Date(b.onDate));
  },
  getPastEvents: state => {
    const today = Date.now();

    return state.eventList
      .filter(e => moment(today).isAfter(e.onDate, "day"))
      .sort((a, b) => new Date(b.onDate) - new Date(a.onDate));
  }
};

const mutations = {
  STORE_EVENT_LIST(state, events) {
    state.eventList = events;
  },
  SET_EVENT_LIST_REFRESH_STATE(state, refreshed) {
    state.eventsRefreshed = refreshed;
  },
  ADD_NEW_EVENT(state, event) {
    state.eventList.push(event);
  },
  UPDATE_EVENT(state, event) {
    state.eventList = state.eventList.map(e => {
      if (e.eventId === event.eventId) return event;
      else return e;
    });
  },
  DELETE_EVENT(state, eventId) {
    state.eventList = state.eventList.filter(e => e.eventId !== eventId);
  }
};

const actions = {
  loadEvents({ commit }) {
    return new Promise((resolve, reject) => {
      commit("SET_EVENT_LIST_REFRESH_STATE", false);
      AxiosCommon.get("event/all")
        .then(result => {
          commit("STORE_EVENT_LIST", result.data);
          commit("SET_EVENT_LIST_REFRESH_STATE", true);
          resolve();
        })
        .catch(err => {
          reject(err);
        });
    });
  },
  addNewEvent({ commit }, eventData) {
    return new Promise((resolve, reject) => {
      AxiosCommon.post("event/create", eventData)
        .then(result => {
          commit("ADD_NEW_EVENT", result.data);
          resolve();
        })
        .catch(err => {
          reject(err);
        });
    });
  },
  updateEvent({ commit }, eventData) {
    return new Promise((resolve, reject) => {
      AxiosCommon.post("event/update", eventData)
        .then(result => {
          commit("UPDATE_EVENT", result.data);
          resolve();
        })
        .catch(err => {
          reject(err);
        });
    });
  },
  deleteEvent({ commit }, eventId) {
    return new Promise((resolve, reject) => {
      AxiosCommon.post(`event/delete/${eventId}`)
        .then(result => {
          commit("DELETE_EVENT", eventId);
          resolve();
        })
        .catch(err => {
          reject(err);
        });
    });
  }
};

export default {
  state,
  getters,
  mutations,
  actions
};

export { staticEventListTypes };
