import AxiosCommon from "../../axios/axios-common";
import moment from "moment";

const staticListTypes = Object.freeze({
  today: {
    type: "TODAY",
    icon: "mdi-calendar-multiselect",
    color: "blue"
  },
  scheduled: {
    type: "SCHEDULED",
    icon: "mdi-clock",
    color: "orange"
  },
  all: {
    type: "ALL",
    icon: "mdi-inbox-full",
    color: "grey"
  },
  flagged: {
    type: "FLAGGED",
    icon: "mdi-flag",
    color: "red"
  }
});

const mappedTodos = Object.freeze({
  type: "MAPPED",
  icon: "mdi-map-marker",
  color: "grey"
});

const customTodos = Object.freeze({
  type: "CUSTOM",
  icon: "mdi-account",
  color: "green"
});

const state = {
  lists: [],
  lastModification: null,
  isListRefreshed: false
};

const getters = {
  getListsNameAndId: state => {
    return state.lists.map(({ name, listId }) => {
      return { name, listId };
    });
  },
  todayTodosCount: state => {
    let count = 0;
    state.lists.forEach(list => {
      count += list.todayTodosCount;
    });

    return count;
  },
  scheduledTodosCount: state => {
    let count = 0;
    state.lists.forEach(list => {
      count += list.scheduledTodosCount;
    });

    return count;
  },
  allTodosCount: state => {
    let count = 0;
    state.lists.forEach(list => {
      count += list.todosCount;
    });

    return count;
  },
  flaggedTodosCount: state => {
    let count = 0;
    state.lists.forEach(list => {
      count += list.flaggedTodosCount;
    });

    return count;
  },
  findTodosByIds: ({ lists }) => ids => {
    const todos = [];

    lists.forEach(list => {
      list.todos.forEach(todo => {
        if (ids.includes(todo.todoId)) todos.push(todo);
        else {
          todo.nestedTodoItems.forEach(nTodo => {
            if (ids.includes(nTodo.todoId)) todos.push(nTodo);
          });
        }
      });
    });

    return todos;
  },
  clientListsView: state => {
    return state.lists.map(list => ({
      listId: list.listId,
      icon: list.icon,
      text: list.name,
      color: list.color,
      group: list.group,
      todosCount: list.todosCount
    }));
  },
  findTodoListByType: ({ lists }) => type => {
    if (!type) return createTodayTodoList(lists);

    switch (type) {
      case staticListTypes.today.type:
        return createTodayTodoList(lists);

      case staticListTypes.all.type:
        return createAllTodoList(lists);

      case staticListTypes.scheduled.type:
        return createScheduledTodoList(lists);

      case staticListTypes.flagged.type:
        return createFlaggedTodoList(lists);

      case mappedTodos.type:
        return createMappedTodoList(lists);

      default:
        return lists.find(list => list.listId === type);
    }
  }
};

const mutations = {
  STORE_TODO_LISTS(state, lists) {
    state.lists = lists;
  },
  SET_LIST_REFRESH_STATE(state, refreshed) {
    state.isListRefreshed = refreshed;
  },
  REFRESH_LAST_MODIFICATION(state) {
    state.lastModification = Date.now();
  },
  ADD_NEW_TODO(state, { todo, listId }) {
    const list = state.lists.find(list => list.listId === listId);

    list.todos.push(todo);
  },
  DELETE_TODO(state, todoD) {
    const list = state.lists.find(list => list.listId === todoD.listId);

    list.todos = list.todos.filter(todo => {
      if (todo.todoId === todoD.todoId) return false;
      else if (todo.nestedTodoItems) {
        todo.nestedTodoItems = todo.nestedTodoItems.filter(
          nestedTodo => nestedTodo.todoId !== todoD.todoId
        );
      }
      return true;
    });
  },
  UPDATE_TODO(state, editedTodo) {
    const list = state.lists.find(list => list.listId === editedTodo.listId);

    list.todos = list.todos.map(todo => {
      if (editedTodo.todoType === "TODO" && todo.todoId === editedTodo.todoId) {
        return editedTodo;
      } else if (editedTodo.todoType === "NESTED_TODO")
        todo.nestedTodoItems = todo.nestedTodoItems.map(nestedTodo => {
          if (nestedTodo.todoId === editedTodo.todoId) return editedTodo;
          else return nestedTodo;
        });
      return todo;
    });
  },
  RECALC_LIST_COUNTS(state, listId) {
    const list = state.lists.find(list => list.listId === listId);

    const today = Date.now();

    let todosCount = 0;
    let allTodosCount = 0;
    let scheduledTodosCount = 0;
    let flaggedTodosCount = 0;
    let todayTodosCount = 0;

    list.todos.forEach(todo => {
      if (!todo.done) {
        todosCount++;
        if (todo.flagged) flaggedTodosCount++;
        if (todo.onDate) {
          scheduledTodosCount++;
          if (moment(todo.onDate).isSame(today, "day")) todayTodosCount++;
        }
      }
      allTodosCount++;

      todo.nestedTodoItems.forEach(nestedTodo => {
        if (!nestedTodo.done) {
          todosCount++;
          if (nestedTodo.flagged) flaggedTodosCount++;
          if (nestedTodo.onDate) {
            scheduledTodosCount++;
            if (moment(nestedTodo.onDate).isSame(today, "day"))
              todayTodosCount++;
          }
        }
        allTodosCount++;
      });
    });

    list.todosCount = todosCount;
    list.allTodosCount = allTodosCount;
    list.scheduledTodosCount = scheduledTodosCount;
    list.flaggedTodosCount = flaggedTodosCount;
    list.todayTodosCount = todayTodosCount;
  }
};

const actions = {
  loadTodoLists({ commit }) {
    return new Promise((resolve, reject) => {
      commit("SET_LIST_REFRESH_STATE", false);
      AxiosCommon.get("todo/list/all")
        .then(result => {
          commit("STORE_TODO_LISTS", result.data);
          commit("REFRESH_LAST_MODIFICATION");
          commit("SET_LIST_REFRESH_STATE", true);
          resolve();
        })
        .catch(err => {
          reject(err);
        });
    });
  },
  addNewListByReferralCode({ commit, dispatch }, code) {
    return new Promise((resolve, reject) => {
      AxiosCommon.post(`todo/list/${code}/add/user`)
        .then(result => {
          dispatch("loadTodoLists");
          resolve();
        })
        .catch(err => {
          reject(err);
        });
    });
  },
  addNewTodo({ commit }, todo) {
    return new Promise((resolve, reject) => {
      AxiosCommon.post("todo/save", todo)
        .then(result => {
          commit("ADD_NEW_TODO", { todo: result.data, listId: todo.listId });
          commit("RECALC_LIST_COUNTS", todo.listId);
          resolve();
        })
        .catch(err => {
          reject(err);
        });
    });
  },
  editTodoList({ commit, dispatch }, listData) {
    commit("SET_LIST_REFRESH_STATE", false);
    return new Promise((resolve, reject) => {
      AxiosCommon.post("todo/list/add", listData)
        .then(result => {
          dispatch("loadTodoLists").then(result => {
            resolve();
          });
        })
        .catch(err => {
          reject(err);
        });
    });
  },
  deleteTodo({ commit }, todo) {
    return new Promise((resolve, reject) => {
      AxiosCommon.post(`todo/delete/${todo.todoId}`)
        .then(result => {
          commit("DELETE_TODO", todo);
          commit("RECALC_LIST_COUNTS", todo.listId);
          resolve();
        })
        .catch(err => {
          reject(err);
        });
    });
  },
  deleteTodoList({ commit, dispatch }, listId) {
    commit("SET_LIST_REFRESH_STATE", false);
    return new Promise((resolve, reject) => {
      AxiosCommon.post(`todo/list/delete/${listId}`)
        .then(result => {
          dispatch("loadTodoLists");
          resolve();
        })
        .catch(err => {
          reject(err);
        });
    });
  },
  editTodo({ commit, dispatch }, { oldListId, editedTodo }) {
    return new Promise((resolve, reject) => {
      if (editedTodo.nestedTodoItems)
        editedTodo.nestedTodoItems.forEach(todo => {
          if (todo.todoId.toString().includes("new")) todo.todoId = null;
        });

      AxiosCommon.post(`todo/edit`, editedTodo)
        .then(result => {
          const newTodo = result.data;
          if (newTodo.listId !== oldListId) {
            commit("ADD_NEW_TODO", { listId: newTodo.listId, todo: newTodo });
            editedTodo.listId = oldListId;
            commit("DELETE_TODO", editedTodo);
            commit("RECALC_LIST_COUNTS", oldListId);
          } else {
            commit("UPDATE_TODO", newTodo);
          }
          commit("RECALC_LIST_COUNTS", newTodo.listId);
          resolve();
        })
        .catch(err => {
          reject(err);
        });
    });
  },
  toggleTodo({ commit, dispatch }, todo) {
    return new Promise((resolve, reject) => {
      AxiosCommon.post(`todo/toggle/${todo.todoId}`)
        .then(result => {
          todo.done = !todo.done;

          if (todo.nestedTodoItems && todo.done)
            todo.nestedTodoItems.forEach(
              nestedTodo => (nestedTodo.done = true)
            );
          commit("RECALC_LIST_COUNTS", todo.listId);
          resolve();
        })
        .catch(err => {
          reject(err);
        });
    });
  },
  findTodos({ commit, getters }, text) {
    return new Promise((resolve, reject) => {
      AxiosCommon.get(`todo/find`, {
        params: {
          text
        }
      })
        .then(result => {
          if (result.data) {
            const todos = getters.findTodosByIds(result.data);

            resolve(completeListByType(todos, customTodos));
            return;
          }

          resolve();
        })
        .catch(err => {
          reject(err);
        });
    });
  },
  deleteWholeList({ commit, dispatch }, listId) {
    commit("SET_LIST_REFRESH_STATE", false);
    return new Promise((resolve, reject) => {
      AxiosCommon.post(`todo/list/delete-whole/${listId}`)
        .then(result => {
          dispatch("loadTodoLists");
          resolve();
        })
        .catch(err => {
          reject(err);
        });
    });
  },
  getListUsers({ commit }, listId) {
    return new Promise((resolve, reject) => {
      AxiosCommon.get(`todo/list/${listId}/all-users`)
        .then(result => {
          resolve(result.data);
        })
        .catch(err => {
          reject(err);
        });
    });
  },
  deleteUserFromList({ commit }, data) {
    return new Promise((resolve, reject) => {
      AxiosCommon.post(`todo/list/${data.listId}/remove-user/${data.userId}`)
        .then(result => {
          resolve();
        })
        .catch(err => {
          reject(err);
        });
    });
  },
  changeListOwner({ dispatch }, data) {
    return new Promise((resolve, reject) => {
      AxiosCommon.post(`todo/list/${data.listId}/change-owner/${data.userId}`)
        .then(result => {
          dispatch("loadTodoLists");
          resolve();
        })
        .catch(err => {
          reject(err);
        });
    });
  }
};

const createTodayTodoList = lists => {
  const today = Date.now();
  const todos = [];

  lists.forEach(list => {
    if (list.todayTodosCount === 0) return;

    list.todos.forEach(todo => {
      if (todo.onDate && moment(todo.onDate).isSame(today, "day"))
        todos.push(todo);
      else
        todo.nestedTodoItems.forEach(nestedTodo => {
          if (
            nestedTodo.onDate &&
            moment(nestedTodo.onDate).isSame(today, "day")
          )
            todos.push(nestedTodo);
        });
    });
  });

  return completeListByType(todos, staticListTypes.today);
};

const createFlaggedTodoList = lists => {
  const todos = [];

  lists.forEach(list => {
    if (list.flaggedTodosCount === 0) return;

    list.todos.forEach(todo => {
      if (!todo.done && todo.flagged) todos.push(todo);
      else
        todo.nestedTodoItems.forEach(nestedTodo => {
          if (!nestedTodo.done && nestedTodo.flagged) todos.push(nestedTodo);
        });
    });
  });

  return completeListByType(todos, staticListTypes.flagged);
};

const createScheduledTodoList = lists => {
  const todos = [];

  lists.forEach(list => {
    if (list.scheduledTodosCount === 0) return;

    list.todos.forEach(todo => {
      if (todo.onDate) todos.push(todo);
      else
        todo.nestedTodoItems.forEach(nestedTodo => {
          if (nestedTodo.onDate) todos.push(nestedTodo);
        });
    });
  });

  return completeListByType(todos, staticListTypes.scheduled);
};

const createAllTodoList = lists => {
  return completeListByType(lists, staticListTypes.all);
};

const createMappedTodoList = lists => {
  const todos = [];

  lists.forEach(list => {
    list.todos.forEach(todo => {
      if (todo.place) todos.push(todo);
      else
        todo.nestedTodoItems.forEach(nestedTodo => {
          if (nestedTodo.place) todos.push(nestedTodo);
        });
    });
  });

  return completeListByType(todos, mappedTodos);
};

const completeListByType = (list, staticList) => {
  switch (staticList.type) {
    case staticListTypes.today.type:
    case staticListTypes.scheduled.type:
    case staticListTypes.flagged.type:
    case mappedTodos.type:
    case customTodos.type:
      return {
        listId: staticList.type,
        icon: staticList.icon,
        color: staticList.color,
        todos: list
      };

    case staticListTypes.all.type:
      return [
        {
          listId: staticList.type,
          icon: staticList.icon,
          color: staticList.color,
          todos: []
        },
        ...list
      ];
    default:
      return {};
  }
};

export default {
  state,
  getters,
  mutations,
  actions
};

export { staticListTypes, mappedTodos, customTodos };
